{% extends "!layout.html" %}

{%- block sidebartitle %}

{%- include "searchbox.html" %}

<a href="https://www.renpy.org/" style="margin-bottom: 0">Ren'Py Visual Novel Engine
<img src="{{ logo_url }}" alt="{{ _('Logo') }}" style="width: 93px; height: 100px; border-radius: 0; margin: 0 auto; padding: 0; margin-top: 5px; background-color: unset;" />
</a>

<a href="{{ pathto(master_doc) }}" style="margin-bottom: 0px">{{ version }} Documentation</a>

{%- endblock %}
