.wy-nav-content {
    max-width: 1080px;
}

@media only screen and (min-width: 481px) {
    [role="search"] {
        top: 12.944px;
        margin-bottom: 12.944px;
        width: 274.125px;
        padding-right: 1em;
        position: fixed;
        z-index: 1;
    }

    .wy-side-nav-search {
        padding-top: 41.944px;
        padding-bottom: 0px !important;
        padding-right: 2.5em !important;
    }

    .wy-nav-side {
        padding-bottom: 0px;
    }

    .wy-side-scroll {
        width: 100%;
    }
}

/* Highlighting */


html[data-theme="light"] .highlight .k, html[data-theme="light"] .highlight .kc {
    color: #4040b0 !important;
    font-weight: normal !important;
}

html[data-theme="light"] .highlight .s1, html[data-theme="light"] .highlight .s2, html[data-theme="light"] .highlight .se, html[data-theme="light"] .highlight .sd {
    color: #935c45 !important;
    font-weight: normal !important;
}

html[data-theme="light"] .highlight .c, html[data-theme="light"] .highlight .c1 {
    color: #008000 !important;
    font-style: normal;
}

html[data-theme="light"] .highlight .mi, html[data-theme="light"] .highlight .mf {
    color: #222 !important;
}

html[data-theme="light"] .highlight .na {
    color: #004040 !important;
}




html[data-theme='dark'] .sig .sig-name .pre, html[data-theme='dark'] .sig .sig-prename .pre  {
    color: #fff !important;
}

html[data-theme='dark'] dl > dt.sig {
    background: #000 !important;
    border-left: 3px solid #444 !important;;
    color: #c0c0c0 !important;
}


html[data-theme='dark'] .highlight .k, html[data-theme='dark'] .highlight .kc {
    color: #c586c0 !important;
    font-weight: normal !important;
}

html[data-theme='dark'] .highlight .s1, html[data-theme='dark'] .highlight .s2, html[data-theme='dark'] .highlight .se, html[data-theme='dark'] .highlight .sd {
    color: #ce9178 !important;
    font-weight: normal !important;
}

html[data-theme='dark'] .highlight .c, html[data-theme='dark'] .highlight .c1 {
    color: #6a9955 !important;
    font-style: normal;
}

html[data-theme='dark'] .highlight .mi, html[data-theme='dark'] .highlight .mf {
    color: #b5cea8 !important;
}

html[data-theme='dark'] .highlight .na {
    color: #9cdcfe !important;
}

html[data-theme='dark'].writer-html5 .rst-content dl[class]:not(.option-list):not(.field-list):not(.footnote):not(.citation):not(.glossary):not(.simple) dl:not(.option-list):not(.field-list):not(.footnote):not(.citation):not(.glossary):not(.simple)>dt {
    background: #0f0f0f;
    color: #959595;
    border-color: #2b2b2b;
}

dt.sig {
    display: block !important;
}

p.caption {
    line-height: inherit !important;
    white-space: normal !important;
    height: auto !important;
    margin: 16px 0 6px 0 !important;
}
